local ffi            = require('ffi')
local json           = require('json')
local msgpack        = require('msgpack')
local json_encode    = json and json.encode
local msgpack_decode = msgpack and msgpack.decode
local ffi_new        = ffi.new
local format, rep    = string.format, string.rep
local insert, remove = table.insert, table.remove
local concat         = table.concat

ffi.cdef([[
struct schema_il_Opcode {
    struct {
        uint16_t op;
        union {
            uint16_t scale;
            uint16_t step;
        };
    };
    union {
        uint32_t ripv;
        uint32_t offset;
        uint32_t name;
        uint32_t len;
    };
    union {
        struct {
            uint32_t ipv;
            uint32_t ipo;
        };
        int32_t  ci;
        int64_t  cl;
        double   cd;
    };

    // block
    // ripv ipv ipo




    /* rt.err_type depends on these values */





};

struct schema_il_V {
    union {
        uint64_t     raw;
        struct {
            uint32_t gen    :30;
            uint32_t islocal:1;
            uint32_t isdead :1;
            uint32_t inc;
        };
    };
};

]])

local opcode = ffi_new('struct schema_il_Opcode')

local op2str = {
    [0xc0   ] = 'CALLFUNC   ',   [0xc1   ] = 'DECLFUNC   ',
    [0xc2    ] = 'IBRANCH    ',   [0xc3    ] = 'SBRANCH    ',
    [0xc4      ] = 'IFSET      ',   [0xc5      ] = 'IFNUL      ',
    [0xc6  ] = 'INTSWITCH  ',   [0xc7  ] = 'STRSWITCH  ',
    [0xc8 ] = 'OBJFOREACH ',   [0xc9       ] = 'MOVE       ',
    [0xca       ] = 'SKIP       ',   [0xcb      ] = 'PSKIP      ',
    [0xcc   ] = 'PUTBOOLC   ',   [0xcd    ] = 'PUTINTC    ',
    [0xce   ] = 'PUTLONGC   ',   [0xcf  ] = 'PUTFLOATC  ',
    [0xd0 ] = 'PUTDOUBLEC ',   [0xd1    ] = 'PUTSTRC    ',
    [0xd2    ] = 'PUTBINC    ',   [0xd3  ] = 'PUTARRAYC  ',
    [0xd4    ] = 'PUTMAPC    ',   [0xd5      ] = 'PUTXC      ',
    [0xd6    ] = 'PUTNULC    ',   [0xd7    ] = 'PUTBOOL    ',
    [0xd8     ] = 'PUTINT     ',   [0xd9    ] = 'PUTLONG    ',
    [0xda   ] = 'PUTFLOAT   ',   [0xdb  ] = 'PUTDOUBLE  ',
    [0xdc     ] = 'PUTSTR     ',   [0xdd     ] = 'PUTBIN     ',
    [0xde   ] = 'PUTARRAY   ',   [0xdf     ] = 'PUTMAP     ',
    [0xe0] = 'PUTINT2LONG',   [0xe1 ] = 'PUTINT2FLT ',
    [0xe2 ] = 'PUTINT2DBL ',   [0xe3] = 'PUTLONG2FLT',
    [0xe4] = 'PUTLONG2DBL',   [0xe5 ] = 'PUTFLT2DBL ',
    [0xe6 ] = 'PUTSTR2BIN ',   [0xe7 ] = 'PUTBIN2STR ',
    [0xe8 ] = 'PUTENUMI2S ',   [0xe9 ] = 'PUTENUMS2I ',
    [0xea     ] = 'ISBOOL     ',   [0xeb      ] = 'ISINT      ',
    [0xee     ] = 'ISLONG     ',   [0xec    ] = 'ISFLOAT    ',
    [0xed   ] = 'ISDOUBLE   ',   [0xef      ] = 'ISSTR      ',
    [0xf0      ] = 'ISBIN      ',   [0xf1    ] = 'ISARRAY    ',
    [0xf2      ] = 'ISMAP      ',   [0xf3      ] = 'ISNUL      ',
    [0xf4 ] = 'ISNULORMAP ',   [0xf5      ] = 'LENIS      ',
    [0xf6      ] = 'ISSET      ',   [0xf7   ] = 'ISNOTSET   ',
    [0xf8   ] = 'BEGINVAR   ',   [0xf9     ] = 'ENDVAR     ',
    [0xfa  ] = 'CHECKOBUF  ',   [0xfb  ] = 'ERRVALUEV  '
}

local function opcode_new(op)
    local o = ffi_new('struct schema_il_Opcode')
    if op then o.op = op end
    return o
end

local opcode_NOP = opcode_new(0xc9); opcode_NOP.ripv = 0xffffffff

local function opcode_ctor_ipv(op)
    return function(ipv)
        local o = ffi_new('struct schema_il_Opcode')
        o.op = op; o.ipv = ipv
        return o
    end
end

local function opcode_ctor_ipv_ipo(op)
    return function(ipv, ipo)
        local o = ffi_new('struct schema_il_Opcode')
        o.op = op; o.ipv = ipv; o.ipo = ipo
        return o
    end
end

local function opcode_ctor_ripv_ipv_ipo(op)
    return function(ripv, ipv, ipo)
        local o = ffi_new('struct schema_il_Opcode')
        o.op = op; o.ripv = ripv or 0xffffffff
        o.ipv = ipv; o.ipo = ipo
        return o
    end
end

local function opcode_ctor_offset_ipv_ipo(op)
    return function(offset, ipv, ipo)
        local o = ffi_new('struct schema_il_Opcode')
        o.op = op; o.offset = offset; o.ipv = ipv; o.ipo = ipo
        return o
    end
end

local function opcode_ctor_offset_ci(op)
    return function(offset, ci)
        local o = ffi_new('struct schema_il_Opcode')
        o.op = op; o.offset = offset; o.ci = ci or 42
        return o
    end
end

local il_methods = {
    nop = function() return opcode_NOP end,
    ----------------------------------------------------------------
    declfunc = function(name, ipv)
        local o = opcode_new(0xc1)
        o.name = name; o.ipv = ipv
        return o
    end,
    ibranch = function(ci)
        local o = opcode_new(0xc2)
        o.ci = ci
        return o
    end,
    ifset       = opcode_ctor_ipv         (0xc4),
    ifnul       = opcode_ctor_ipv_ipo     (0xc5),
    intswitch   = opcode_ctor_ipv_ipo     (0xc6),
    strswitch   = opcode_ctor_ipv_ipo     (0xc7),
    objforeach  = opcode_ctor_ripv_ipv_ipo(0xc8),
    ----------------------------------------------------------------
    move        = opcode_ctor_ripv_ipv_ipo(0xc9),
    skip        = opcode_ctor_ripv_ipv_ipo(0xca),
    pskip       = opcode_ctor_ripv_ipv_ipo(0xcb),
    ----------------------------------------------------------------
    putboolc    = function(offset, cb)
        local o = opcode_new(0xcc)
        o.offset = offset; o.ci = cb and 1 or 0
        return o
    end,
    putintc     = opcode_ctor_offset_ci(0xcd),
    putlongc = function(offset, cl)
        local o = opcode_new(0xce)
        o.offset = offset; o.cl = cl
        return o
    end,
    putfloatc   = function(offset, cf)
        local o = opcode_new(0xcf)
        o.offset = offset; o.cd = cf
        return o
    end,
    putdoublec  = function(offset, cd)
        local o = opcode_new(0xd0)
        o.offset = offset; o.cd = cd
        return o
    end,
    putarrayc   = opcode_ctor_offset_ci(0xd3),
    putmapc     = opcode_ctor_offset_ci(0xd4),
    putnulc = function(offset)
        local o = opcode_new(0xd6)
        o.offset = offset
        return o
    end,
    ----------------------------------------------------------------
    putbool     = opcode_ctor_offset_ipv_ipo(0xd7),
    putint      = opcode_ctor_offset_ipv_ipo(0xd8),
    putlong     = opcode_ctor_offset_ipv_ipo(0xd9),
    putfloat    = opcode_ctor_offset_ipv_ipo(0xda),
    putdouble   = opcode_ctor_offset_ipv_ipo(0xdb),
    putstr      = opcode_ctor_offset_ipv_ipo(0xdc),
    putbin      = opcode_ctor_offset_ipv_ipo(0xdd),
    putarray    = opcode_ctor_offset_ipv_ipo(0xde),
    putmap      = opcode_ctor_offset_ipv_ipo(0xdf),
    putint2long = opcode_ctor_offset_ipv_ipo(0xe0),
    putint2flt  = opcode_ctor_offset_ipv_ipo(0xe1),
    putint2dbl  = opcode_ctor_offset_ipv_ipo(0xe2),
    putlong2flt = opcode_ctor_offset_ipv_ipo(0xe3),
    putlong2dbl = opcode_ctor_offset_ipv_ipo(0xe4),
    putflt2dbl  = opcode_ctor_offset_ipv_ipo(0xe5),
    putstr2bin  = opcode_ctor_offset_ipv_ipo(0xe6),
    putbin2str  = opcode_ctor_offset_ipv_ipo(0xe7),
    ----------------------------------------------------------------
    isbool      = opcode_ctor_ipv_ipo(0xea),
    isint       = opcode_ctor_ipv_ipo(0xeb),
    islong      = opcode_ctor_ipv_ipo(0xee),
    isfloat     = opcode_ctor_ipv_ipo(0xec),
    isdouble    = opcode_ctor_ipv_ipo(0xed),
    isstr       = opcode_ctor_ipv_ipo(0xef),
    isbin       = opcode_ctor_ipv_ipo(0xf0),
    isarray     = opcode_ctor_ipv_ipo(0xf1),
    ismap       = opcode_ctor_ipv_ipo(0xf2),
    isnul       = opcode_ctor_ipv_ipo(0xf3),
    isnulormap  = opcode_ctor_ipv_ipo(0xf4),
    ----------------------------------------------------------------
    lenis = function(ipv, ipo, len)
        local o = opcode_new(0xf5)
        o.ipv = ipv; o.ipo = ipo; o.len = len
        return o
    end,
    ----------------------------------------------------------------
    isnotset    = opcode_ctor_ipv(0xf7),
    beginvar    = opcode_ctor_ipv(0xf8),
    endvar      = opcode_ctor_ipv(0xf9),
    ----------------------------------------------------------------
    checkobuf = function(offset, ipv, ipo, scale)
        local o = opcode_new(0xfa)
        o.offset = offset; o.ipv = ipv or 0xffffffff
        o.ipo = ipo or 0; o.scale = scale or 1
        return o
    end,
    errvaluev  = opcode_ctor_ipv_ipo(0xfb)
    ----------------------------------------------------------------
    -- callfunc, sbranch, putstrc, putbinc, putxc and isset
    -- are instance methods
}

-- visualize register
local function rvis(reg, inc)
    if reg == 0xffffffff then
        return '_'
    elseif not inc or inc == 0 then
        return '$'..reg
    else
        return format('$%d+%d', reg, inc)
    end
end

-- visualize constant
local function cvis(o, extra, decode)
    if extra then
        local c = extra[o]
        if c then
            local ok, res = true, c
            if decode then
                ok, res = pcall(decode, c)
            end
            if ok then
                ok, res = pcall(json_encode, res)
            end
            if ok then
                return res
            end
        end
    end
    return format('#%p', o)
end

-- visualize opcode
local function opcode_vis(o, extra)
    local opname = op2str[o.op]
    if o == opcode_NOP then
        return 'NOP'
    elseif o.op == 0xc0 then
        return format('%s %s,\t%s,\tFUNC<%s>', opname,
                      rvis(o.ripv), rvis(o.ipv, o.ipo), extra[o])
    elseif o.op == 0xc1 then
        return format('%s %d,\t%s', opname, o.name, rvis(o.ipv))
    elseif o.op == 0xc2 then
        return format('%s %d', opname, o.ci)
    elseif o.op == 0xc3 then
        return format('%s %s', opname, cvis(o, extra))
    elseif o.op == 0xc4 or (
           o.op >= 0xf7 and o.op <= 0xf9) then
        return format('%s %s', opname, rvis(o.ipv))
    elseif (o.op >= 0xc5 and o.op <= 0xc7) or
           (o.op >= 0xea and o.op <= 0xf4) or
           o.op == 0xfb then
        return format('%s [%s]', opname, rvis(o.ipv, o.ipo))
    elseif o.op == 0xc8 then
        return format('%s %s,\t[%s],\t%d', opname, rvis(o.ripv), rvis(o.ipv, o.ipo), o.step)
    elseif o.op == 0xc9 then
        return format('%s %s,\t%s', opname, rvis(o.ripv), rvis(o.ipv, o.ipo))
    elseif o.op == 0xf6 then
        return format('%s %s,\t%s,\t%s', opname, rvis(o.ripv), rvis(o.ipv, o.ipo), cvis(o, extra))
    elseif o.op == 0xca or o.op == 0xcb then
        return format('%s %s,\t[%s]', opname, rvis(o.ripv), rvis(o.ipv, o.ipo))
    elseif o.op == 0xcc or o.op == 0xcd or
           o.op == 0xd3 or o.op == 0xd4 then
        return format('%s [%s],\t%d', opname, rvis(0, o.offset), o.ci)
    elseif o.op == 0xce then
        return format('%s [%s],\t%s', opname, rvis(0, o.offset), o.cl)
    elseif o.op == 0xcf or o.op == 0xd0 then
        return format('%s [%s],\t%f', opname, rvis(0, o.offset), o.cd)
    elseif o.op == 0xd6 then
        return format('%s [%s]', opname, rvis(0, o.offset))
    elseif o.op == 0xd1 or o.op == 0xd2 then
        return format('%s [%s],\t%s', opname, rvis(0, o.offset), cvis(o, extra))
    elseif o.op == 0xd5 then
        return format('%s [%s],\t%s', opname, rvis(0, o.offset), cvis(o, extra, msgpack_decode))
    elseif o.op >= 0xd7 and o.op <= 0xe7 then
        return format('%s [%s],\t[%s]', opname, rvis(0, o.offset), rvis(o.ipv, o.ipo))
    elseif o.op == 0xe8 or o.op == 0xe9 then
        return format('%s [%s],\t[%s],\t%s', opname,
                      rvis(0, o.offset), rvis(o.ipv, o.ipo), cvis(o, extra))
    elseif o.op == 0xf5 then
        return format('%s [%s],\t%d', opname, rvis(o.ipv, o.ipo), o.len)
    elseif o.op == 0xfa then
        return format('%s %s,\t[%s],\t%d', opname, rvis(0, o.offset), rvis(o.ipv, o.ipo), o.scale)
    else
        return format('<opcode: %d>', o.op)
    end
end

-- visualize IL code
local il_vis_helper
il_vis_helper = function(res, il, indentcache, level, object)
    if not object then -- pass
    elseif type(object) == 'table' then
        local start = 1
        if level ~= 0 or type(object[1]) ~= 'table' then
            il_vis_helper(res, il, indentcache, level, object[1])
            start = 2
            level = level + 1
        end
        for i = start, #object do
            il_vis_helper(res, il, indentcache, level, object[i])
        end
    else
        local indent = indentcache[level]
        if not indent then
            indent = '\n'..rep('  ', level)
            indentcache[level] = indent
        end
        insert(res, indent)
        insert(res, il.opcode_vis(object))
    end
end
local function il_vis(il, root)
    local res = {}
    il_vis_helper(res, il, {}, 0, root)
    if res[1] == '\n' then res[1] = '' end
    insert(res, '\n')
    return concat(res)
end

-- cleanup IL:
--  (*) fix incorrect lists nesting
--  (*) remove NOPS and MOVEs assigning to NILREG
--  (*) remove redundant IFs
local il_cleanup_helper
il_cleanup_helper = function(res, object)
    if type(object) == 'table' then
        local head = object[1]
        if type(head) == 'cdata' and
           head.op >= 0xc1 and head.op <= 0xc8 then
            local nested = { head }
            for i = 2, #object do
                il_cleanup_helper(nested, object[i])
            end
            if nested[2] or head.op == 0xc1 then
                insert(res, nested) -- keep nested block IFF it's not empty
            end
        else
            for i = 1, #object do
                il_cleanup_helper(res, object[i])
            end
        end
    elseif object.op < 0xc9 or object.op > 0xcb or
           object.ripv ~= 0xffffffff then
        insert(res, object) -- elide NOPs (NOP is MOVE to NILREG)
    end
end
local function il_cleanup(object)
    local res = {}
    il_cleanup_helper(res, object)
    return res
end

-- === Basic optimization engine in less than 450 LOC. ===
--
-- Elide some MOVE $reg, $reg+offset instructions.
-- Combine COBs (CHECKOBUFs) and hoist them out of loops.
--
-- We track variable state in a 'scope' data structure.
-- A scope is a dict keyed by a variable name. Scopes are chained
-- via 'parent' key. The lookup starts with the outter-most scope
-- and walks the chain until the entry is found.
--
-- The root scope is created upon entering DECLFUNC. When optimiser
-- encounters a nested block, it adds another scope on top of the chain
-- and invokes itself recursively. Once the block is done, the child
-- scope is removed.
--
-- When a variable value is updated in a program (e.g. assignment),
-- the updated entry is created in the outter-most scope, shadowing
-- any entry in parent scopes. Once a block is complete, the outermost
-- scope has all variable changes captured. These changes are merged
-- with the parent scope.
--
-- Variable state is <gen, inc, islocal, isdead> tuple. Variables are
-- scoped lexically; if a variable is marked local it is skipped when
-- merging with the parent scope. Codegen may explicitly shorten variable
-- lifetime with a ENDVAR instruction. Once a variable is dead its value
-- is no longer relevant. Valid code MUST never reference a dead variable.
--
-- Note: ENDVAR affects the code following the instruction (in a depth-first
--       tree walk order.) E.g. if the first branch in a conditional
--       declares a variable dead, it affects subsequent branches as well.
--
-- Concerning gen and inc.
--
-- Gen is a generation counter. If a variable X is updated
-- in such a way that new(X) = old(X) + C, the instruction is elided
-- and C is added to inc while gen is unchanged. Those we call 'dirty'
-- variables. If a variable is updated in a different fashion, including
-- but not limited to copying another variable's value, i.e. X = Y, we
-- declare that a new value is unrelated to the old one. Gen is assigned
-- a new unique id, and inc is set to 0.
--
-- Surprisingly, this simple framework is powerful enough to determine
-- if a loop proceeds in fixed increments (capture variable state
-- at the begining of the loop body and at the end and compare gen-s.)

local function vlookup(scope, vid)
    if not scope then
        assert(false, format('$%d not found', vid)) -- assume correct code
    end
    local res = scope[vid]
    if res then
        if res.isdead == 1 then
            assert(false, format('$%d is dead', vid)) -- assume correct code
        end
        return res
    end
    return vlookup(scope.parent, vid)
end

local function vcreate(scope, vid)
    local v = scope[vid]
    if not v then
        v = ffi_new('struct schema_il_V')
        scope[vid] = v
    end
    return v
end

local function vsnapshot(scope, vid)
    local v = scope[vid]
    if v then
        local copy = ffi_new('struct schema_il_V')
        copy.raw = v.raw
        return copy
    else -- in outter scope, virtually immutable
        return vlookup(scope.parent, vid)
    end
end

-- 'execute' an instruction and update scope
local function vexecute(il, scope, o, res)
    assert(type(o)=='cdata')
    if o.op == 0xc9 and o.ripv == o.ipv then
       local vinfo = vlookup(scope, o.ipv)
       local vnewinfo = vcreate(scope, o.ipv)
       vnewinfo.gen = vinfo.gen
       vnewinfo.inc = vinfo.inc + o.ipo
       return
    end
    if o.op == 0xf8 then
        local vinfo = vcreate(scope, o.ipv)
        vinfo.islocal = 1
        insert(res, o)
        return
    end
    if o.op == 0xf9 then
        local vinfo = vcreate(scope, o.ipv)
        vinfo.isdead = 1
        insert(res, o)
        return
    end
    local fixipo = 0
    if (o.op == 0xc0 or
        o.op >= 0xc5 and o.op <= 0xcb or
        o.op >= 0xd7 and o.op <= 0xf6 or
        o.op == 0xfa or o.op == 0xfb) and
       o.ipv ~= 0xffffffff then

        local vinfo = vlookup(scope, o.ipv)
        fixipo = vinfo.inc
    end
    local fixoffset = 0
    if o.op >= 0xcc and o.op <= 0xe9 or
       o.op == 0xfa then

        local vinfo = vlookup(scope, 0)
        fixoffset = vinfo.inc
    end
    -- spill $0
    if o.op == 0xc0 then
        local vinfo = vlookup(scope, 0)
        if vinfo.inc > 0 then
            insert(res, il.move(0, 0, vinfo.inc))
        end
        vinfo = vcreate(scope, 0)
        vinfo.gen = il.id()
        vinfo.inc = 0
    end
    -- apply fixes
    o.ipo = o.ipo + fixipo
    o.offset = o.offset + fixoffset
    insert(res, o)
    if (o.op == 0xc0 or
        o.op >= 0xc8 and o.op <= 0xcb) and
       o.ripv ~= 0xffffffff then
        local vinfo = vcreate(scope, o.ripv)
        vinfo.gen = il.id()
        vinfo.inc = 0
    end
end

-- Merge branches (conditional or switch);
-- if branches had diverged irt $r, spill $r
-- (e.g. one did $1 = $1 + 2 while another didn't touch $1 at all)
-- Note: bscopes[1]/bblocks[1] are unused, indices start with 2.
local function vmergebranches(il, bscopes, bblocks)
    local parent = bscopes[2].parent
    local skip = { parent = true } -- vids to skip: dead or diverged
    local maybediverged = {}
    local diverged = {}
    local counters = {} -- how many branches have it, if < #total then diverged
    local nscopes = #bscopes
    for i = 2, nscopes do
        for vid, vinfo in pairs(bscopes[i]) do
            counters[vid] = (counters[vid] or 1) + 1
            if vinfo.islocal == 0 and not skip[vid] then
                local vother = maybediverged[vid]
                if vinfo.isdead == 1 then
                    skip[vid] = true
                    diverged[vid] = nil
                    maybediverged[vid] = nil
                    vcreate(parent, vid).isdead = 1 
                elseif not vother then
                    maybediverged[vid] = vinfo
                elseif vother.raw ~= vinfo.raw then
                    skip[vid] = true
                    maybediverged[vid] = nil
                    diverged[vid] = true
                end
            end
        end
    end
    for vid, vinfo in pairs(maybediverged) do
        if counters[vid] ~= nscopes then
            diverged[vid] = true
        else
            local vpinfo = vcreate(parent, vid)
            vpinfo.gen = vinfo.gen
            vpinfo.inc = vinfo.inc
        end
    end
    for vid, _ in pairs(diverged) do
        local vpinfo = vcreate(parent, vid)
        vpinfo.gen = il.id()
        vpinfo.inc = 0
        for i = 2, nscopes do
            local vinfo = bscopes[i][vid]
            if vinfo and vinfo.inc > 0 then
                insert(bblocks[i], il.move(vid, vid, vinfo.inc))
            end
        end
    end
    return diverged
end

-- Collect variables modified by the given code block.
local vvarschanged_cache = setmetatable({}, { __mode = 'k' })
local vvarschanged
vvarschanged = function(block)
    local res = vvarschanged_cache[block]
    if res then return res end
    assert(type(block) == 'table')
    res = {}
    local locals = {}
    for i = 1, #block do
        local item = block[i]
        if type(item) == 'table' then
            local nested = vvarschanged(item)
            for vid, _ in pairs(nested) do
                res[vid] = true
            end
        elseif item.op == 0xc0 then
            res[0] = true
            if item.ripv ~= 0xffffffff then res[item.ripv] = true end
        elseif item.op >= 0xc8 and item.op <= 0xcb then
            res[item.ripv] = true
        elseif item.op == 0xf8 then
            locals[item.ipv] = true
        end
    end
    for vid, _ in pairs(locals) do
        res[vid] = nil
    end
    vvarschanged_cache[block] = res
    return res
end

-- Spill 'dirty' variables that are modified in a loop body
-- before entering the loop.
local function vprepareloop(il, scope, lblock, block)
    local changed = vvarschanged(lblock)
    for vid, _ in pairs(changed) do
        local vinfo = vlookup(scope, vid)
        if vinfo.inc > 0 then
            insert(block, il.move(vid, vid, vinfo.inc))
            local info = vcreate(scope, vid)
            info.gen = il.id()
            info.inc = 0
        end
    end
end

-- Spill 'dirty' variables at the end of a loop body.
local function vmergeloop(il, lscope, lblock)
    local parent = lscope.parent
    for vid, vinfo in pairs(lscope) do
        if vinfo ~= parent and vinfo.islocal == 0 then
            local vpinfo = vcreate(parent, vid)
            vpinfo.gen = il.id()
            vpinfo.inc = 0
            if vinfo.inc > 0 then
                insert(lblock, il.move(vid, vid, vinfo.inc))
            end
        end
    end
end

-- Whether or not it is valid to move a COB across the specified
-- code range. If start is non-nil block[start] is assumed to be
-- another COB we are attempting to merge with.
local vcobmotionvalid
vcobmotionvalid = function(block, cob, start, stop)
    if cob.ipv == 0xffffffff then
        return true
    else
        local t = block[start]
        if t and t.ipv ~= 0xffffffff then
            return false -- they won't merge
        end
    end
    for i = start or 1, stop or #block do
        local o = block[i]
        if type(o) == 'table' then
            if not vcobmotionvalid(o, cob) then
                return false
            end
        elseif o.op == 0xf2 or o.op == 0xf1 then
            -- we lack alias analysis; it's unsafe to move past ANY typecheck
            return false
        elseif o.op >= 0xc8 and o.op <= 0xcb and
               o.ripv == cob.ipv then
            -- it modifies the variable
            return false
        end
    end
    return true
end

-- Merge 2 COBs, update a.
local function vcobmerge(a, b)
    if a.offset < b.offset then
        a.offset = b.offset
    end
    if b.ipv ~= 0xffffffff then
        assert(a.ipv == 0xffffffff)
        a.ipv = b.ipv
        a.ipo = b.ipo
        a.scale = b.scale
    end
end

-- Here be dragons.
local voptimizeblock
voptimizeblock = function(il, scope, block, res)
    -- COB hoisting state
    local entryv0sn = vsnapshot(scope, 0) -- $0 at block start
    local firstcobpos, firstcobv0sn -- first COB (if any), will attempt
                                    -- to pop it into the parent
    local cobpos, cobv0sn, cobfixup -- 'active' COB, when we encounter
                                    -- another COB, we attempt to merge

    for i = 2, #block do -- foreach item in the current block, excl. head
        local o = block[i]
        if type(o) == 'cdata' then -- Opcode
            vexecute(il, scope, o, res)
            if o.op == 0xfa then
                local v0info = vlookup(scope, 0)
                if not cobv0sn or v0info.gen ~= cobv0sn.gen or
                   not vcobmotionvalid(res, o, cobpos) then
                    -- no active COB or merge imposible: activate current COB
                    cobpos = #res
                    cobv0sn = vsnapshot(scope, 0)
                    if not firstcobpos then
                        firstcobpos = cobpos
                        firstcobv0sn = cobv0sn 
                    end
                    cobfixup = 0
                else
                    -- update active COB and drop the current one
                    o.offset = o.offset + cobfixup
                    vcobmerge(o, res[cobpos])
                    res[cobpos] = o
                    remove(res)
                end
            end
        else
            local head = o[1]
            if head.op >= 0xc4 and head.op <= 0xc7 then
                -- branchy things: conditions and switches
                local bscopes = {0}
                local bblocks = {head}
                local cobhoistable, cobmaxoffset = 0, 0
                for i = 2, #o do
                    local branch = o[i]
                    local bscope = { parent = scope }
                    local bblock = { branch[1] }
                    voptimizeblock(il, bscope, branch, bblock)
                    bscopes[i] = bscope
                    bblocks[i] = bblock
                    -- update hoistable COBs counter
                    local cob = bblock[0]
                    if cob and cob.ipv == 0xffffffff then
                        cobhoistable = cobhoistable + 1
                        if cob.offset > cobmaxoffset then
                            cobmaxoffset = cob.offset
                        end
                    end
                end
                -- a condition has 2 branches, though empty ones are omitted;
                -- if it's the case temporary restore the second branch
                -- for the vmergebranches() to consider this execution path
                if #o == 2 and (head.op == 0xc4 or
                                head.op == 0xc5) then
                    bscopes[3] = { parent = scope }
                end
                vmergebranches(il, bscopes, bblocks)
                insert(res, bblocks)
                -- hoist COBs but only if at least half of the branches will
                -- benefit
                if cobhoistable >= #bblocks/2 then
                    for i = 2, #bblocks do
                        local bblock = bblocks[i]
                        local pos = bblock[-1]
                        if pos and bblock[0].ipv == 0xffffffff then
                            remove(bblock, pos)
                        end
                    end
                    local o = il.checkobuf(cobmaxoffset)
                    local v0info = vlookup(scope, 0)
                    if not cobv0sn or v0info.gen ~= cobv0sn.gen or
                       not vcobmotionvalid(res, o, cobpos) then
                        -- no active COB or merge imposible: activate current
                        cobpos = #res
                        cobv0sn = vsnapshot(scope, 0)
                        if not firstcobpos then
                            firstcobpos = cobpos
                            firstcobv0sn = cobv0sn 
                        end
                        cobfixup = 0
                        insert(res, cobpos, o)
                    else
                        -- update active COB and drop the current one
                        o.offset = o.offset + cobfixup
                        vcobmerge(o, res[cobpos])
                        res[cobpos] = o
                    end
                end
            elseif head.op == 0xc8 then
                -- loops
                local v0bsn = vsnapshot(scope, 0) -- used by COB hoisting
                vprepareloop(il, scope, o, res)
                local v0asn = vsnapshot(scope, 0) -- used by COB hoisting
                local lscope = { parent = scope }
                local lblock = {}
                vexecute(il, lscope, head, lblock)
                local ivinfo = lscope[head.ripv]
                local ivgen = ivinfo.gen
                voptimizeblock(il, lscope, o, lblock)
                if ivinfo.gen == ivgen then
                    -- loop variable incremented in fixed steps
                    head.step = ivinfo.inc
                    lscope[head.ripv] = nil
                    local vinfo = vcreate(scope, head.ripv)
                    vinfo.gen = il.id()
                    vinfo.inc = 0
                end
                vmergeloop(il, lscope, lblock)
                insert(res, lblock)
                local cob = lblock[0]
                if cob then -- attempt COB hoisting
                    local v0info = vlookup(lscope, 0)
                    if v0info.gen ~= v0asn.gen or
                       v0info.inc == v0asn.inc then
                        -- hoisting failed: $0 is non-linear or
                        -- doesn't change at all
                    else
                        remove(lblock, lblock[-1])
                        cob.offset = 0
                        cob.ipv = head.ipv
                        cob.ipo = head.ipo
                        cob.scale = v0info.inc --[[ - v0asn.inc
                            but the later == 0, vprepareloop() spilled it ]]
                        if not cobv0sn or v0bsn.gen ~= cobv0sn.gen or
                           not vcobmotionvalid(res, cob, cobpos) then
                            -- no active COB or merge imposible:
                            -- activate current one
                            cobfixup = 0
                            cobpos = #res
                            cobv0sn = vsnapshot(scope, 0)
                            if not firstcobpos then
                                firstcobpos = cobpos
                                firstcobv0sn = cobv0sn
                            end
                            insert(res, cobpos, cob)
                        else
                            -- update active COB and drop the current one;
                            -- changing cobv0sn here is a hack
                            -- enabling to move COBs across (some) loops.
                            -- Basically when moving COB we must compensate
                            -- for $0 changing. Due to limitations in the
                            -- variable state tracking, it looks as if $0
                            -- value before and after the loop are unrelated.
                            -- In fact, after($0) = before($0) + k*x.
                            cobfixup = v0bsn.inc - cobv0sn.inc
                            cob.offset = cobfixup
                            vcobmerge(cob, res[cobpos])
                            cobv0sn = vsnapshot(scope, 0)
                            res[cobpos] = cob
                        end
                    end
                end
            else
                assert(false)
            end
        end
    end
    -- Add missing ENDVARs.
    for vid, vinfo in pairs(scope) do
        if vid ~= 'parent' and vinfo.islocal == 1 and vinfo.isdead == 0 then
            insert(res, il.endvar(vid))
        end
    end
    -- Attempt to pop the very first COB into the parent.
    if firstcobpos and firstcobv0sn.gen == entryv0sn.gen and
       vcobmotionvalid(res, res[firstcobpos], nil, firstcobpos) then
        -- There was a COB and the code motion was valid.
        -- Create a copy of the COB with adjusted offset and save it in res[0].
        -- If the parent decides to accept, it has to remove a now redundant
        -- COB at firstcobpos.
        local o = res[firstcobpos]
        res[0] = il.checkobuf(o.offset + firstcobv0sn.inc - entryv0sn.inc,
                               o.ipv, o.ipo, o.scale)
        res[-1] = firstcobpos
    end
end

local function voptimizefunc(il, func)
    local head, scope = func[1], {}
    local res = { head }
    local r0 = vcreate(scope, 0)
    local r1 = vcreate(scope, head.ipv)
    voptimizeblock(il, scope, func, res)
    if r0.inc > 0 then
        insert(res, il.move(0, 0, r0.inc))
    end
    if r1.inc > 0 then
        insert(res, il.move(head.ipv, head.ipv, r1.inc))
    end
    return res
end

local function voptimize(il, code)
    local res = {}
    for i = 1, #code do
        res[i] = voptimizefunc(il, code[i])
    end
    return res
end

local function il_create()

    local extra = {}
    local id = 10 -- low ids are reserved

    local il
    il = setmetatable({
        callfunc = function(ripv, ipv, ipo, func)
            local o = opcode_new(0xc0)
            o.ripv = ripv or 0xffffffff; o.ipv = ipv; o.ipo = ipo
            extra[o] = func
            return o
        end,
        ----------------------------------------------------------------
        sbranch = function(cs)
            local o = opcode_new(0xc3)
            extra[o] = cs
            return o
        end,
        putstrc = function(offset, cs)
            local o = opcode_new(0xd1)
            o.offset = offset; extra[o] = cs
            return o
        end,
        putbinc = function(offset, cb)
            local o = opcode_new(0xd2)
            o.offset = offset; extra[o] = cb
            return o
        end,
        putxc = function(offset, cx)
            local o = opcode_new(0xd5)
            o.offset = offset; extra[o] = cx
            return o
        end,
        putenums2i = function(offset, ipv, ipo, tab)
            local o = opcode_new(0xe9)
            o.offset = offset; o.ipv = ipv; o.ipo = ipo
            extra[o] = tab
            return o
        end,
        putenumi2s = function(offset, ipv, ipo, tab)
            local o = opcode_new(0xe8)
            o.offset = offset; o.ipv = ipv; o.ipo = ipo
            extra[o] = tab
            return o
        end,
        isset = function(ripv, ipv, ipo, cs)
            local o = opcode_new(0xf6)
            o.ripv = ripv; o.ipv = ipv
            o.ipo = ipo; extra[o] = cs
            return o
        end,
    ----------------------------------------------------------------
        id = function() id = id + 1; return id end,
        get_extra = function(o)
            return extra[o]
        end,
        vis = function(code) return il_vis(il, code) end,
        opcode_vis = function(o)
            return opcode_vis(o, extra)
        end,
        cleanup = il_cleanup,
        optimize = function(code) return voptimize(il, code) end,
    }, { __index = il_methods })
    return il
end

return {
    il_create = il_create
}
